// --------------------------------------------------------------------------------------------------------------------
// <copyright file="CommandLineHandler.cs" company="Matthias Friedrich">
//   Copyright  Matthias Friedrich 2013
// </copyright>
// --------------------------------------------------------------------------------------------------------------------

namespace CommandlineHandling
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Reflection;

    public class CommandLineHandler
    {
        private readonly IDictionary<string, Type> knownArgumentTypes = new Dictionary<string, Type>();

        public CommandLineHandler()
        {
            this.Mapper = OptionMapper.DefaultMapper;
        }

        public ICommandLineOptionMapper Mapper { get; set; }

        public Func<string, ICommandLineOption> ResolveOption { get; set; }

        public Func<string, Type> ResolveOptionType { get; set; }

        /// <summary>
        /// Returns true, if the given type is a valid option type. A valid option type implements the ICommandLineOption-interface.
        /// </summary>
        public static bool IsOptionType(Type t)
        {
            string interfaceName = typeof(ICommandLineOption).FullName;

            return t.GetInterface(interfaceName) != null;
        }

        /// <summary>
        /// Runs all executable commands.
        /// </summary>
        public void Execute(IEnumerable<ICommandLineOption> arguments)
        {
            ExceptionExtensions.ThrowArgumentNullException(arguments, () => arguments);

            foreach (ICommandLineOption arg in arguments)
            {
                arg.InvokeExecute();
            }
        }

        /// <summary>
        /// Parses the given arguments and executes all commands.
        /// </summary>
        public IEnumerable<ICommandLineOption> Parse(IEnumerable<string> args)
        {
            ExceptionExtensions.ThrowArgumentNullException(args, () => args);

            var parser = new CommandLineParser(this);

            IEnumerable<ICommandLineOption> parsedOptions = parser.Parse(args);

            ICommandLineOptionMapper mapper;
            if ((mapper = this.Mapper) == null)
            {
                return parsedOptions;
            }
            else
            {
                List<ICommandLineOption> options = parsedOptions.ToList();
                foreach (ICommandLineOption next in options)
                {
                    mapper.MapObject(next);
                }

                return options.AsReadOnly();
            }
        }

        /// <summary>
        /// Registers all argument types provided by the given assembly.
        /// </summary>
        public void Register(Assembly assembly)
        {
            ExceptionExtensions.ThrowArgumentNullException(assembly, () => assembly);

            IEnumerable<ProvideCommandLineOptionAttribute> providedOptions = GetProvidedOptionsFromAssembly(assembly);

            foreach (ProvideCommandLineOptionAttribute option in providedOptions)
            {
                string optionName = GetOptionName(option);
                if (!string.IsNullOrEmpty(optionName))
                {
                    this.knownArgumentTypes[optionName] = option.ArgumentType;
                }
            }
        }

        public void Register<T>(string name) where T : class, ICommandLineOption
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            this.knownArgumentTypes[name] = typeof(T);
        }

        public void Register<T>() where T : class, ICommandLineOption
        {
            IEnumerable<OptionAttribute> optionAttributes = typeof(T).ReflectAttributes<OptionAttribute>(true);
            foreach (OptionAttribute next in optionAttributes)
            {
                this.knownArgumentTypes[next.Name] = typeof(T);
            }
        }

        /// <summary>
        /// Unregisters all option types provided by the given assembly.
        /// </summary>
        public void Unregister(Assembly assembly)
        {
            ExceptionExtensions.ThrowArgumentNullException(assembly, () => assembly);

            IEnumerable<ProvideCommandLineOptionAttribute> providedOptions = GetProvidedOptionsFromAssembly(assembly);
            foreach (ProvideCommandLineOptionAttribute option in providedOptions)
            {
                string optionName = GetOptionName(option);
                if (!string.IsNullOrEmpty(optionName))
                {
                    Unregister(optionName);
                }
            }
        }

        public void Unregister(string name)
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            if (this.knownArgumentTypes.ContainsKey(name))
            {
                this.knownArgumentTypes.Remove(name);
            }
        }

        public void UnregisterAll()
        {
            this.knownArgumentTypes.Clear();
        }

        /// <summary>
        /// Returns the name for the given option.
        /// </summary>
        private static string GetOptionName(ProvideCommandLineOptionAttribute option)
        {
            Type t = option.ArgumentType;

            if (IsOptionType(t))
            {
                string optionName = option.Name;

                OptionAttribute a;
                if ((a = t.ReflectAttribute<OptionAttribute>(true).FirstOrDefault()) != null)
                {
                    optionName = a.Name;
                }

                return optionName;
            }

            return null;
        }

        /// <summary>
        /// Returns a list of options provided by the given assembly.
        /// </summary>
        private static IEnumerable<ProvideCommandLineOptionAttribute> GetProvidedOptionsFromAssembly(Assembly assembly)
        {
            return assembly.GetCustomAttributes(typeof(ProvideCommandLineOptionAttribute), true)
                .Select<object, ProvideCommandLineOptionAttribute>((x) => (ProvideCommandLineOptionAttribute)x);
        }

        protected internal ICommandLineOption Resolve(string name)
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            Func<string, ICommandLineOption> resolveFunc = this.ResolveOption;
            if (resolveFunc != null)
            {
                return resolveFunc(name);
            }

            return null;
        }

        protected internal Type ResolveType(string name)
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            if (this.knownArgumentTypes.ContainsKey(name))
            {
                return this.knownArgumentTypes[name];
            }
            else
            {
                Func<string, Type> resolveFunc = this.ResolveOptionType;
                if (resolveFunc != null)
                {
                    return resolveFunc(name);
                }
            }

            return typeof(Option);
        }
    }
}