// --------------------------------------------------------------------------------------------------------------------
// <copyright file="Option.cs" company="Matthias Friedrich">
//   Copyright  Matthias Friedrich 2013
// </copyright>
// --------------------------------------------------------------------------------------------------------------------

namespace CommandlineHandling
{
    using System;
    using System.Collections.Generic;
    using System.Linq;

    public class Option : ICommandLineOption
    {
        private IList<OptionParameter> parameters = new List<OptionParameter>();

        public Option()
        {
        }

        public string Name { get; set; }

        string ICommandLineOption.Value { get; set; }

        public Func<ICommandLineOption, ICommandLineOption, object> Execute { get; set; }

        public IList<OptionParameter> Parameters
        {
            get { return this.parameters; }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                this.parameters = value;
            }
        }

        public OptionParameter Parameter(string name, bool throwException = false)
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            OptionParameter p = this.parameters.SingleOrDefault(x => x.Name == name);

            if ((p == null) && throwException)
            {
                throw new MissingOptionException();
            }

            return p;
        }

        public IEnumerable<OptionParameter> ParameterList(string name)
        {
            ExceptionExtensions.ThrowArgumentNullException(name, () => name);

            return this.parameters.Where((x) => x.Name == name);
        }

        public ICommandLineOption Next { get; set; }

        public T Find<T>(bool directFollowerOnly = true) where T : class, ICommandLineOption
        {
            if (this.Next == null)
            {
                return default(T);
            }

            T next;
            if ((next = this.Next as T) != null)
            {
                return next;
            }

            if (directFollowerOnly)
            {
                return default(T);
            }

            return this.Next.Find<T>();
        }

        public IEnumerable<T> FindAll<T>(bool directFollowerOnly = true) where T : class, ICommandLineOption
        {
            var all = new List<T>();

            var next = this.Find<T>(directFollowerOnly);

            while (next != null)
            {
                all.Add(next);
                next = this.Find<T>(directFollowerOnly);
            }

            return all.AsReadOnly();
        }

        object ICommandLineOption.InvokeExecute(ICommandLineOption caller)
        {
            Func<ICommandLineOption, ICommandLineOption, object> handler = this.Execute;
            if (handler != null)
            {
                return handler(caller, this);
            }

            return null;
        }

        protected string TryGetParameter(string name, bool throwExeception)
        {
            OptionParameter p = this.Parameter(name);
            if (p != null)
            {
                return p.Value;
            }

            if (throwExeception)
            {
                ThrowMissingArgumentException(name);
            }

            return null;
        }

        private static void ThrowMissingArgumentException(string name)
        {
            throw new MissingOptionException(name);
        }

        #region Operators

        public static explicit operator bool(Option p)
        {
            return p.GetValue<bool>();
        }

        public static explicit operator short(Option p)
        {
            return p.GetValue<short>();
        }

        public static explicit operator char(Option p)
        {
            return p.GetValue<char>();
        }

        public static explicit operator int(Option p)
        {
            return p.GetValue<int>();
        }

        public static explicit operator long(Option p)
        {
            return p.GetValue<long>();
        }

        public static explicit operator byte(Option p)
        {
            return p.GetValue<byte>();
        }

        public static explicit operator float(Option p)
        {
            return p.GetValue<float>();
        }

        public static explicit operator double(Option p)
        {
            return p.GetValue<double>();
        }

        public static explicit operator decimal(Option p)
        {
            return p.GetValue<decimal>();
        }

        public static explicit operator DateTime(Option p)
        {
            return p.GetValue<DateTime>();
        }

        #endregion
    }

    public abstract class Option<T> : Option
    {
        public T Value
        {
            get { return this.GetValue<T>(); }
        }
    }
}